# include <stdio.h>
# include <stdlib.h>
# include <strings.h>
# include <ctype.h>
# include <errno.h>
# include <syslog.h>

# define EXTERN extern
# include "slow.h"


void
milterr(char *prog, char *file, int line, int log, int err, char *what, char *why)
{
	if (isdebug(BUG_SOURCE))
	{
	    if (log == FALSE)
		(void)fprintf(stderr, "%s: %s(line %d): %s: %s: %s\n",
			prog==NULL?"milter":prog,
			file==NULL?"":file,
			line,
			what==NULL?"":what,
			why==NULL?"":why,
			err==0?"Warning":strerror(err));
	    else
		(void)syslog(LOG_INFO, "%s: %s(line %d): %s: %s: %s",
			prog==NULL?"milter":prog,
			file==NULL?"":file,
			line,
			what==NULL?"":what,
			why==NULL?"":why,
			err==0?"Warning":strerror(err));
	}
	else
	{
	    if (log == FALSE)
		(void)fprintf(stderr, "%s: %s: %s: %s\n",
			prog==NULL?"milter":prog,
			what==NULL?"":what,
			why==NULL?"":why,
			err==0?"Warning":strerror(err));
	    else
		(void)syslog(LOG_INFO, "%s: %s: %s: %s",
			prog==NULL?"milter":prog,
			what==NULL?"":what,
			why==NULL?"":why,
			err==0?"Warning":strerror(err));
	}
	if (err == 0)
		exit(EAGAIN);
	else
		exit(err);
}

void
print_list(char *title, char **list, FILE *fp)
{
	char **l;

	if (list == NULL || fp == NULL)
		return;

	for (l = list; *l != NULL; ++l)
		(void) fprintf(fp, "%s: %s\n", 
			title==NULL?"<nil>":title, *l);
	return;
}

char **
free_list(char **list)
{
	char **l;

	if (list == NULL)
		return NULL;

	for (l = list; *l != NULL; ++l)
		(void) free(*l);
	(void) free(list);
	return NULL;
}

char **
push_list(char *name, char **list)
{
	char **l;
	int num;

	if (name == NULL)
		return list;

	if (isdebug(BUG_PUSHN))
	{ 
		(void) printf("push_list(%s):\n", name);
	}

	if (list == NULL)
	{
		list = try_alloc(__FILE__, __LINE__,
			sizeof(char *) * 2);
		list[0] = try_strdup(__FILE__, __LINE__, name);
		list[1] = NULL;
		if (isdebug(BUG_PUSHN))
		{ 
			(void) printf("  added\t%s\n", list[0]);
			(void) printf("\tNULL\n");
		}
		return list;
	}

	num = 0;
	for (l = list; *l != NULL; ++l)
	{
		if (isdebug(BUG_PUSHN))
		{ 
			(void) printf("\t%s\n", *l);
		}
		++num;
	}

	list = try_realloc(__FILE__, __LINE__, 
		list, sizeof(char *) * (num + 2));
	list[num++] = try_strdup(__FILE__, __LINE__, name);
	list[num] = NULL;
	if (isdebug(BUG_PUSHN))
	{ 
		(void) printf("  added\t%s\n", list[num-1]);
		(void) printf("\tNULL\n");
	}
	return list;
}


void
read_honey(void)
{
	FILE *fp;
	char buf[BUFSIZ];
	char *cp;
	int nhosts;

	if (HoneyHosts != NULL)
	{
		milterr(Prog, __FILE__, __LINE__, FALSE,
			0, HoneyFile, "Already initialized");
	}

	fp = fopen(HoneyFile, "r");
	if (fp == NULL)
	{
		milterr(Prog, __FILE__, __LINE__, FALSE,
			errno, "Open to read", HoneyFile);
	}
	nhosts = 0;
	while (fgets(buf, BUFSIZ, fp) != NULL)
	{
		if (*buf == '#' || *buf == '\n')
			continue;
		if ((cp = strchr(buf, '\n')) != NULL)
			*cp = '\0';
		HoneyHosts = push_list(buf, HoneyHosts);
	}
	if (ferror(fp))
	{
		milterr(Prog, __FILE__, __LINE__, FALSE,
			errno, "While reading", HoneyFile);
	}
	(void) fclose(fp);
	return;
}

bool
is_in_list(char *what, char **list)
{
	char **h;
	int len;

	if (what == NULL || list == NULL)
		return FALSE;
	if ((len = strlen(what)) == 0)
		return FALSE;

	for (h = list; *h != NULL; ++h)
	{
		if (strncasecmp(what, *h, len) == 0)
			return TRUE;
	}
	return FALSE;
}


typedef struct debugcat {
	char *name;
	unsigned int  bit;
} DEBUGCAT;

static DEBUGCAT Debugcats[] = {
	{"mx",			BUG_MX},
	{"db",			BUG_DB},
	{"push_list",		BUG_PUSHN},
	{"honey",		BUG_HONEY},
	{"source",		BUG_SOURCE},
	{"connect",		BUG_CONNECT},
	{"helo",		BUG_HELO},
	{"mailfrom",		BUG_MAILFROM},
	{"rcptto",		BUG_RCPTTO},
	{"header",		BUG_HEADER},
	{"eoh",			BUG_EOH},
	{"eom",			BUG_EOM},
	{"close",		BUG_CLOSE},
	{"abort",		BUG_ABORT},
	{NULL,			0},
};

void
set_debug(char *catagory)
{
	DEBUGCAT *d;
	int len;

	if (catagory == NULL)
		return;
	if ((len = strlen(catagory)) == 0)
		return;

	for (d = Debugcats; d->name != NULL; ++d)
	{
		if (strncasecmp(catagory, d->name, len) == 0)
		{
			setbit(d->bit, Debugbits);
			return;
		}
	}
	(void) fprintf(stderr, "-D%s: unknown debugging catagory\n", catagory);
	exit (EPROTO);
	return;
}


char **
mkargv(char *string, int delim)
{
	char **argv;
	char *cp, *sp;
	int save;

	if (string == NULL)
		return NULL;
	argv = NULL;
	if (delim == 0)
	{
		return (push_list(cleanup(string), argv));
	}
	for (cp = string; *cp != '\0'; ++cp)
	{
		if (delim == ' ')
			sp = strpbrk(cp, " \t\r\n");
		else
			sp = strchr(cp, delim);
		if (sp != NULL)
		{
			save = *sp;
			*sp = '\0';
			argv = push_list(cleanup(cp), argv);
			*sp = save;
			cp = sp;
			continue;
		}
		return (push_list(cleanup(cp), argv));
	}
	return argv;
}

char *
cleanup(char *string)
{
	char *cp, *ep;

	for (cp = string; ;++cp)
	{
		if (! isspace((int)*cp))
			break;
	}
	for (ep = cp + strlen(cp) -1; ; --ep)
	{
		if (!isspace((int)*ep))
			break;
	}
	if (ep <= cp)
		return cp;
	*(ep +1) = '\0';
	return cp;
}

/*
** Bullet proof (or let's say more resilient)
** memory allocation routines.
*/
void *
try_alloc(char *file, int line, size_t bytes)
{
	void *r = NULL;
	int cnt = 0;

	if (bytes <= 0)
		bytes = 1;

	while (r == NULL)
	{
		if ((r = malloc(bytes)) != NULL)
			break;
		(void)sleep(++cnt);
		if (cnt >= MAX_FAILS_M)
		{
			milterr(Prog, file, line, TRUE, errno,
				"ABORTED", "Allocate memory");
		}
	}
	return r;
}

void *
try_strdup(char *file, int line, char *string)
{
	void *r;

	if (string == NULL)
		return NULL;

	r = try_alloc(file, line, strlen(string) +1);
	(void)strcpy(r, string);
	return r;
}

void *
try_realloc(char *file, int line, void *ptr, size_t len)
{
	void *r = NULL;
	int cnt = 0;

	if (ptr == NULL)
		return NULL;
	if (len <= 0)
		len = 1;

	while (r == NULL)
	{
		if ((r = realloc(ptr, len)) != NULL)
			break;
		(void)sleep(++cnt);
		if (cnt >= MAX_FAILS_M)
		{
			milterr(Prog, file, line, TRUE, errno,
				"ABORTED", "Allocate memory");
		}
	}
	return r;
}
